import sys

whitespace = '\t\n\v\f\r '

class Unexpected:
    def __init__(self, file, line, column, value):
        self.file = file
        self.line = line
        self.column = column
        self.value = value
    def __repr__(self):
        return 'Unexpected(%r, %r, %r, %r)' % (self.file, self.line, self.column, self.value)
    def __str__(self):
        return "%s at %s:%d:%d" % (self.value, self.file, self.line, self.column)

class TrackingStream:
    """ Wrapper for input streams that tracks line/column

        mutable field unexpecting - if set to any "True"-like value,
            raise an Unexpected with the position and that value (e.g. a string)
        field name
        fields line and column track the position of get()
    """
    def __init__(self, name = None, input = None):
        if name is None and input is None:
            self.name = '/dev/stdin'
            self._input = sys.stdin
        elif input is None:
            self.name = name
            self._input = open(name)
        else:
            self.name = name
            self._input = input
        self.unexpecting = False

        # Actually sets self._cached to the first character in the stream
        # and self.line and self.column to 1
        self._cached = '\n'
        self.line = 0
        self.advance()

    # C++: ++it
    def advance(self):
        if not self._cached: # EOF
#            print "not advancing " + self.name + " - EOF!"
            return
        if self._cached == '\n':
            self.line += 1
            self.column = 1
        elif self._cached == '\t':
            # Careful with this!
            # Python's modulus returns the same sign as the divisor
            # C++11 modulus returns the same sign as the dividend
            # We have a negative dividend and a positive divisor, and want a positive result.
            self.column += (-self.column)%8 + 1
        else:
            self.column += 1

        old_cached = self._cached
        self._cached = self._input.read(1)
#        print 'advancing %s from %r to %r' % (self.name, old_cached, self._cached)

    # C++: *it
    def get(self):
        if not self._cached:
            if self.unexpecting:
                raise Unexpected(self.name, self.line, self.column, self.unexpecting)
        return self._cached

    # C++: *++it
    def advance_and_get(self):
        self.advance()
        return self.get()
    # C++: *it++ (but is this supported for input iterators?)
    # Hm ... first copy the iterator, including the cached value, then increment 'this'
    # That should work ...
    def get_and_advance(self):
        s = self.get()
        self.advance()
        return s

# Lexer: BeginList, EndList, String, Token, Comment
# Parser: List, Int, String, Token, [Comment]

class List: # Only from parser, not lexer
    def __init__(self, line, column, value):
        self.line = line
        self.column = column
        self.value = value

    def __repr__(self):
        return 'sexpr.List(%d, %d, %r)' % (self.line, self.column, self.value)

    def __str__(self):
        return '(' + ' '.join(map(str, self.value)) + ')'

class BeginList:
    def __init__(self, line, column):
        self.line = line
        self.column = column

class EndList:
    def __init__(self, line, column):
        self.line = line
        self.column = column

class Comment:
    """ Will be ignored by most parsers.

    #.*\n
    //.*\n
    /\*.*\*/
    """
    def __init__(self, line, column, value):
        self.line = line
        self.column = column
        self.value = value

    def __repr__(self):
        return "sexpr.Comment(%d, %d, %r)" % (self.line, self.column, self.value)

    def __str__(self):
        return self.value

def escape(ch, for_string):
    if for_string is False and ch == ' ':
        return "\\ "
    if ch in "'\"\\":
        return "\\" + ch
    if ch >= ' ' and ch <= '~':
        return ch
    if ch == '\n':
        return "\\n"
    if ch == '\r':
        return "\\r"
    if ch == '\t':
        return "\\t"
    if ch == '\a':
        return "\\a"
    if ch == '\b':
        return "\\b"
    if ch == '\x1b':
        return "\\e"
    if ch == '\v':
        return "\\v"
    if ch == '\f':
        return "\\f"
    return "\\x%02x" % ord(ch)

class String: # Stripped of quotes - it may be empty
    def __init__(self, line, column, value):
        self.line = line
        self.column = column
        self.value = value

    def __repr__(self):
        return "sexpr.String(%d, %d, %r)" % (self.line, self.column, self.value)

    def __str__(self):
        out = '"'
        for ch in self.value:
            out += escape(ch, True)
        return out + '"'

class Int: # Only from parser, not lexer
    def __init__(self, line, column, value):
        self.line = line
        self.column = column
        self.value = value

    def __repr__(self):
        return 'sexpr.Int(%d, %d, %d)' % (self.line, self.column, self.value)

    def __str__(self):
        return str(self.value)

    # Will this be representable in an int64_t ?
    def okay(self):
        return -0x8000000000000000 <= self.value and self.value <= 0x7fffffffffffffff

class Token: # Includes ints when returned from Lexer; never empty
    def __init__(self, line, column, value):
        self.line = line
        self.column = column
        self.value = value

    def __repr__(self):
        return "sexpr.Token(%d, %d, %r)" % (self.line, self.column, self.value)

    def __str__(self):
        out = ''
        for ch in self.value:
            out += escape(ch, False)
        return out

    def maybe_int(self):
        try:
            # Note: int() may return a python long
            return Int(self.line, self.column, int(self.value, 0))
        except ValueError:
            return self

class Lexer:
    def __init__(self, name_or_stream = None, input = None):
        if isinstance(name_or_stream, TrackingStream):
            self.source = name_or_stream
        else:
            self.source = TrackingStream(name_or_stream, input)
        self.depth = [] # Positions of unmatched ( so far
    def __iter__(self):
        return self
    def next(self):
        self.source.unexpecting = False
        # First, skip whitespace
        while True: # Whitespace skipper
            line = self.source.line
            column = self.source.column
            ch = self.source.get_and_advance()
#            print 'get: ' + repr(ch)
            if not ch:
                if self.depth:
                    loc = self.depth[-1]
                    raise Unexpected(self.source.name, loc[0], loc[1], "unmatched '('")
                else:
                    raise StopIteration
            if not ch in whitespace:
                break
        # Whitespace skipper


        if ch == '(':
            self.depth.append((line, column))
            return BeginList(line, column)
        if ch == ')':
            if not self.depth:
                raise Unexpected(self.source.name, line, column, "unmatched ')'")
            self.depth.pop()
            return EndList(line, column)
        if ch == '"':
            str = '' # Quotes are not part of a string literal
            self.source.unexpecting = "EOF in string literal"
            while True:
                ch = self.source.get_and_advance()
                if ch == '"':
                    # Quotes are not part of a string literal
                    return String(line, column, str)
                if ch != '\\':
                    str += ch
                    continue
                self.source.unexpecting = "EOF in backslash sequence in string literal"
                str += self._read_after_backslash()
                self.source.unexpecting = "EOF in string literal"
            # while True
        # if ch == '"'

#        print "Beginning token or comment"
        # to get here, it's either a token (including integers) or a comment
        # Note: unexpecting is NOT set
        tok = ''
        if ch == '#':
#            print "Returning hash comment"
            return Comment(line, column, self._read_until_with_prefix('#', '\n'))
        if ch == '/':
            ch = self.source.get()
            if ch in whitespace: # Including ''
                return Token(line, column, '/')
            if ch in '(")':
                print "Warning: some parsers may have difficulty with a '" + ch + "' attached to a token"
                return Token(line, column, tok)
            self.source.advance()
            if ch == '/':
                return Comment(line, column, self._read_until_with_prefix('//', '\n'))
            if ch == '*':
                self.source.unexpecting = "EOF in /* comme"
                return Comment(line, column, self._read_until_with_prefix('/*', '*/'))
            # else, just an ordinary token
            tok = '/'
        # if ch == '/'

        # Still have one in ch; tok may be '' or '/'
        if ch != '\\':
            tok += ch
        else:
            self.source.unexpecting = "EOF in backslash sequence in token"
            tok += self._read_after_backslash()
            self.source.unexpecting = False
        
        # read until whitespace
        while True:
            ch = self.source.get()
            # '' is always in whitespace. This is as it should be.
            if ch in whitespace:
                return Token(line, column, tok)
            if ch in '(")':
                print "Warning: some parsers may have difficulty with a '" + ch + "' attached to a token"
                return Token(line, column, tok)
            self.source.advance()
            # Note: comment characters to appear in a token silently
            if ch != '\\':
                tok += ch
            else: # ch == '\\'
                self.source.unexpecting = "EOF in backslash sequence in token"
                tok += self._read_after_backslash()
                self.source.unexpecting = False
        # while True
    # next()

    def _read_after_backslash(self):
        ch = self.source.get_and_advance()
        if ch in whitespace:
            return ch
        if ch in '"\'\\':
            return ch
        if ch == 'n':
            return '\n'
        if ch == 'r':
            return '\r'
        if ch == 't':
            return '\t'
        if ch == 'a':
            return '\a'
        if ch == 'b':
            return '\b'
        if ch == 'e':
            return '\x1b'
        if ch == 'v':
            return '\v'
        if ch == 'f':
            return '\f'
        if ch == 'x':
            ch1 = self.source.get_and_advance()
            ch2 = self.source.get_and_advance()
            return chr(ch1 + ch2)
        if ch in '01234567':
            oct_str = ch
            ch2 = self.source.get()
            if ch2 in '01234567':
                self.source.advance()
                oct_str += ch2
                ch3 = self.source.get()
                if ch3 in '01234567':
                    self.source.advance()
                    oct_str += ch3
                # else ch3 will be read again from the outer loop by self.source.get_and_advance
            # else ch2 will be read again from the outer loop by self.source.get_and_advance
            return chr(int(oct_str, 8))
        raise Unexpected(self.source.name, self.source.line, self.source.column, "Character following backslash: " + repr(ch))
    def _read_until_with_prefix(self, str, sfx):
        sl = len(sfx)
        while True:
            new_str = self.source.get_and_advance()
            # If EOF was a bad thing, the caller would set .unexpecting
            if not new_str:
                return str
            str += new_str
            if str[-sl:] == sfx:
                return str

class Parser:
    """ Takes the lexer and makes data

        There are two significant things that happen here:
        1. Turn BeginList(), ..., EndList() into a List
        2. Turn some Token into Int
    """
    def __init__(self, lexer, comments):
        self.lexer = lexer
        self.ignore_comments = not comments

    def __iter__(self):
        return self

    def nxt(self):
        pass

    def next(self):
        # The lexer already takes care of ( vs ) matching
        lx = self.lexer.next()
        if self.ignore_comments:
            while True:
                if not isinstance(lx, Comment):
                    break
                lx = self.lexer.next()
        if not isinstance(lx, BeginList):
            if isinstance(lx, Token):
                return lx.maybe_int()
            return lx
        line = lx.line
        column = lx.column
        list = []
        while True:
            # Not self.lexer.next() - we recurse
            lx = self.next()
            if isinstance(lx, EndList):
                return List(line, column, list)
            list.append(lx)
        

# Use python syntax to build SExprs
# unfortunately you can't omit the commas :P

# TODO: see if there's a way to steal meaningful line/column info.
def python_to_sexpr(expr):
    # It's easy to see how this might happen
    # Note that this is the only way to embed a comment or token
    # It would probably be possible (but definitely evil) to implement
    # some sort of syntax extension using pdb, see the 'goto' module?
    for c in List, Int, String, Token, Comment:
        if isinstance(expr, c):
            return expr
    if isinstance(expr, str):
        return String(0, 0, expr)
    if isinstance(expr, int) or isinstance(expr, long):
        return Int(0, 0, expr)
    # No tokens or comments at this level
    list = []
    for elt in expr:
        list.append(python_to_sexpr(elt))
    return List(0, 0, list)

